package GUT;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.type.TypeKind;

import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.AnnotatedArrayType;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.types.AnnotatedTypeMirror.AnnotatedExecutableType;
import checkers.types.AnnotatedTypeMirror.AnnotatedTypeVariable;
import checkers.types.BasicAnnotatedTypeFactory;
import checkers.types.TreeAnnotator;
import checkers.types.TypeAnnotator;
import checkers.util.Pair;
import checkers.util.TreeUtils;

import com.sun.source.tree.BinaryTree;
import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.CompoundAssignmentTree;
import com.sun.source.tree.ExpressionTree;
import com.sun.source.tree.IdentifierTree;
import com.sun.source.tree.MemberSelectTree;
import com.sun.source.tree.MethodInvocationTree;
import com.sun.source.tree.NewArrayTree;
import com.sun.source.tree.NewClassTree;
import com.sun.source.tree.Tree;
import com.sun.source.tree.Tree.Kind;
import com.sun.source.tree.TypeCastTree;
import com.sun.source.tree.UnaryTree;

/**
 * Apply viewpoint adaptation and add implicit annotations to "this" and
 * "super".
 *
 * @author wmdietl
 */
public class GUTAnnotatedTypeFactory extends
BasicAnnotatedTypeFactory<GUTChecker> {

    public GUTAnnotatedTypeFactory(GUTChecker checker, CompilationUnitTree root) {
        // disable flow checker, as it seems to ignore the results of
        // postAsMemberOf
        super(checker, root, false);

        addAliasedAnnotation(org.jmlspecs.annotation.Peer.class, checker.PEER);
        addAliasedAnnotation(org.jmlspecs.annotation.Rep.class, checker.REP);
        addAliasedAnnotation(org.jmlspecs.annotation.Readonly.class, checker.ANY);
        // see GUTVisitor.isPure for handling of pure
    }

    /**
     * The type of "this" is always "self".
     */
    @Override
    public AnnotatedDeclaredType getSelfType(Tree tree) {
        AnnotatedDeclaredType type = super.getSelfType(tree);
        if (type != null) {
            type.clearAnnotations();
            type.addAnnotation(checker.SELF);
        }
        return type;
    }

    @Override
    public boolean canHaveAnnotatedTypeParameters() {
        // In GUT, just allow annotations on the bounds.
        return false;
    }

    /**
     * Adapt the type of field accesses.
     */
    @Override
    protected void postAsMemberOf(AnnotatedTypeMirror type,
            AnnotatedTypeMirror owner, Element element) {
        assert type != null;
        assert owner != null;
        assert element != null;

        if (type.getKind() != TypeKind.DECLARED
                && type.getKind() != TypeKind.ARRAY) {
            // nothing to do
            return;
        }
        if (element.getKind() == ElementKind.LOCAL_VARIABLE
                || element.getKind() == ElementKind.PARAMETER) {
            // the type of local variables and parameters also do not need to
            // change
            return;
        }

        // TODO: why is this fromElement and not getAnnotatedType??
        AnnotatedTypeMirror decltype = this.fromElement(element);

        /*
          System.out.println("\npostasmemberof:");
          System.out.println("in type: " + type);
          System.out.println("decl type: " + decltype);
          System.out.println("owner: " + owner);
          System.out.println("element: " + element);
         */

        // Combine annotations
        AnnotatedTypeMirror combinedType = GUTQualifierUtils.combineTypeWithType(owner, decltype);
        // System.out.println("combined type: " + combinedType);

        // Replace annotations
        type.clearAnnotations();
        type.addAnnotation(GUTQualifierUtils.getUniverseModifier(combinedType));
        // System.out.println("result: " + type.toString());

        // TODO: HACK: we needed to make setTypeArguments public to work
        // around a limitation of the framework.
        // If the method had a return value, like constructor/methodFromUse, we
        // could probably get around this.

        if (type.getKind() == TypeKind.DECLARED
                && combinedType.getKind() == TypeKind.DECLARED) {
            AnnotatedDeclaredType declaredType = (AnnotatedDeclaredType) type;
            AnnotatedDeclaredType declaredCombinedType = (AnnotatedDeclaredType) combinedType;
            declaredType.setTypeArguments(declaredCombinedType.getTypeArguments());
        } else if (type.getKind() == TypeKind.ARRAY
                && combinedType.getKind() == TypeKind.ARRAY) {
            AnnotatedArrayType arrayType = (AnnotatedArrayType) type;
            AnnotatedArrayType arrayCombinedType = (AnnotatedArrayType) combinedType;
            arrayType.setComponentType(arrayCombinedType.getComponentType());
        }

        // System.out.println("out type: " + type);
    }


    @Override
    public List<AnnotatedTypeVariable> typeVariablesFromUse(
            AnnotatedDeclaredType type, TypeElement element) {

        List<AnnotatedTypeVariable> orig = super.typeVariablesFromUse(type, element);
        List<AnnotatedTypeVariable> res = new LinkedList<AnnotatedTypeVariable>();
        AnnotationMirror firstMainModifier = GUTQualifierUtils.getUniverseModifier(type);

        for (AnnotatedTypeVariable atv : orig) {
            AnnotatedTypeVariable  newatv =
                (AnnotatedTypeVariable) GUTQualifierUtils.combineModifierWithType(firstMainModifier, atv);
            res.add(newatv);
        }
        return res;
    }

    /*
     * TODO: unify constructorFromUse and methodFromUse to share more code.
     */

    /**
     * Adapt the types of constructors.
     *
     * @param tree
     *            the new class tree.
     * @return the modified constructor.
     */
    @Override
    public Pair<AnnotatedExecutableType, List<AnnotatedTypeMirror>> constructorFromUse(NewClassTree tree) {
        assert tree != null;

        // using super would substitute too much
        // AnnotatedExecutableType constructor = super.constructorFromUse(tree);

        ExecutableElement ctrElt = TreeUtils.elementFromUse(tree);
        // TODO: why is this fromElement and not getAnnotatedType??
        AnnotatedExecutableType constructor = this.fromElement(ctrElt);

        Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mappings = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

        // Get the result type
        AnnotatedTypeMirror resultType = getAnnotatedType(tree);

        // Modify parameters
        for (AnnotatedTypeMirror parameterType : constructor.getParameterTypes()) {
            AnnotatedTypeMirror combinedType = GUTQualifierUtils.combineTypeWithType(resultType, parameterType);
            mappings.put(parameterType, combinedType);
        }

        // TODO: upper bounds, throws?
        /*for (AnnotatedTypeVariable v : constructor.getTypeVariables() ) {
            System.out.println("v: " + v);
          }*/

        constructor = constructor.substitute(mappings);

        // determine substitution for constructor type variables
        Map<AnnotatedTypeVariable, AnnotatedTypeMirror> typeVarMapping = atypes.findTypeArguments(tree);
        List<AnnotatedTypeMirror> typeargs = new LinkedList<AnnotatedTypeMirror>();

        if (!typeVarMapping.isEmpty()) {
            for ( AnnotatedTypeVariable tv : constructor.getTypeVariables()) {
                typeargs.add(typeVarMapping.get(tv));
            }

            constructor = constructor.substitute(typeVarMapping);
        }

        // System.out.println("adapted constructor: " + constructor);

        return Pair.of(constructor, typeargs);
    }

    /**
     * Adapt the types of methods.
     *
     * @param tree
     *            the method invocation tree.
     * @return the modified method invocation.
     */
    @Override
    public Pair<AnnotatedExecutableType, List<AnnotatedTypeMirror>> methodFromUse(MethodInvocationTree tree) {
        assert tree != null;

        // Calling super would already substitute type variables and doesn't
        // work!
        // AnnotatedExecutableType method = super.methodFromUse(tree);

        ExecutableElement methodElt = TreeUtils.elementFromUse(tree);
        // TODO: why is this fromElement and not getAnnotatedType??
        AnnotatedExecutableType method = this.fromElement(methodElt);

        // System.out.println("Declared method: " + method);

        Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mappings = new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

        // Set the receiver
        AnnotatedTypeMirror receiverType = null;
        ExpressionTree exprTree = tree.getMethodSelect();
        if (exprTree.getKind() == Kind.MEMBER_SELECT) {
            MemberSelectTree memberSelectTree = (MemberSelectTree) exprTree;
            receiverType = getAnnotatedType(memberSelectTree.getExpression());
        } else {
            receiverType = getSelfType(tree);
            // is the following needed? remove!
            // receiverType.clearAnnotations();
            // receiverType.addAnnotation(GUTChecker.SELF);
        }
        assert receiverType != null;
        // System.out.println("Receiver: " + receiverType);

        // Modify parameters
        for (AnnotatedTypeMirror parameterType : method.getParameterTypes()) {
            // System.out.println("Method parameter in: " + parameterType);
            AnnotatedTypeMirror combined = GUTQualifierUtils.combineTypeWithType(receiverType, parameterType);
            mappings.put(parameterType, combined);
            // System.out.println("Method parameter out: " + combined);
        }

        // Modify return type
        AnnotatedTypeMirror returnType = getAnnotatedType(method.getElement()).getReturnType();
        if (returnType.getKind() != TypeKind.VOID) {
            AnnotatedTypeMirror combinedType = GUTQualifierUtils.combineTypeWithType(receiverType, returnType);

            // Careful: we need to use the original return type for the mapping, otherwise the substitution doesn't work
            // TODO: might this be a problem anywhere else? Why is this happening here?
            AnnotatedTypeMirror origRet = method.getReturnType();
            mappings.put(origRet, combinedType);
        }

        // TODO: upper bounds, throws?

        method = method.substitute(mappings);

        // determine substitution for method type variables
        Map<AnnotatedTypeVariable, AnnotatedTypeMirror> typeVarMapping = atypes.findTypeArguments(tree);
        List<AnnotatedTypeMirror> typeargs = new LinkedList<AnnotatedTypeMirror>();

        if (!typeVarMapping.isEmpty()) {
            for ( AnnotatedTypeVariable tv : method.getTypeVariables()) {
                typeargs.add(typeVarMapping.get(tv));
            }

            method = method.substitute(typeVarMapping);
        }

        // System.out.println("adapted method: " + method);

        return Pair.of(method, typeargs);
    }

    /**
     * Create our own TreeAnnotator.
     *
     * @param checker
     *            the GUTChecker.
     * @return the new TreeAnnotator.
     */
    @Override
    protected TreeAnnotator createTreeAnnotator(GUTChecker checker) {
        assert checker != null;
        return new GUTTreeAnnotator(checker);
    }

    /**
     * Create our own TypeAnnotator.
     *
     * @param checker
     *            the GUTChecker.
     * @return the new TreeAnnotator.
     */
    @Override
    protected TypeAnnotator createTypeAnnotator(GUTChecker checker) {
        assert checker != null;
        return new GUTTypeAnnotator(checker);
    }

    /**
     * Currently only needed to add the "self" modifier to "super".
     */
    private class GUTTreeAnnotator extends TreeAnnotator {

        private GUTTreeAnnotator(GUTChecker checker) {
            super(checker, GUTAnnotatedTypeFactory.this);
        }

        /**
         * Add @Self to "super" identifier.
         *
         * @param node
         *            the identifier node.
         * @param p
         *            the AnnotatedTypeMirror to modify.
         */
        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror p) {
            assert node != null;
            assert p != null;

            // There's no "super" kind, so make a string comparison.
            if (node.getName().contentEquals("super")) {
                p.clearAnnotations();
                p.addAnnotation(checker.SELF);
            }

            return super.visitIdentifier(node, p);
        }

        @Override
        public Void visitBinary(BinaryTree node, AnnotatedTypeMirror p) {
            // System.out.println("Binary: ");
            return super.visitBinary(node, p);
        }

        @Override
        public Void visitUnary(UnaryTree node, AnnotatedTypeMirror p) {
            // System.out.println("Unary: ");
            return super.visitUnary(node, p);
        }

        @Override
        public Void visitCompoundAssignment(CompoundAssignmentTree node,
                AnnotatedTypeMirror p) {
            // System.out.println("Compound: ");
            return super.visitCompoundAssignment(node, p);
        }

        @Override
        public Void visitTypeCast(TypeCastTree node, AnnotatedTypeMirror p) {
            // System.out.println("Cast: ");
            return super.visitTypeCast(node, p);
        }

        @Override
        public Void visitNewArray(NewArrayTree tree, AnnotatedTypeMirror type) {
            // System.out.println("Array type: " + type);
            super.visitNewArray(tree, type);
            // ((AnnotatedArrayType)type).getComponentType().addAnnotation(GUTChecker.BOTTOM);
            // System.out.println("Final Array type: " + type);
            return null;
        }

    }

    /**
     */
    private static class GUTTypeAnnotator extends TypeAnnotator {

        public GUTTypeAnnotator(GUTChecker checker) {
            super(checker);
            this.gchecker = checker;
        }
        private final GUTChecker gchecker;

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, ElementKind p) {
            if (!type.isAnnotated()) {
                if (GUTChecker.isAnyDefault(type)) {
                    type.addAnnotation(gchecker.ANY);
                } else {
                    type.addAnnotation(gchecker.PEER);
                }
            }
            return super.visitDeclared(type, p);
        }

        @Override
        public Void visitArray(AnnotatedArrayType type, ElementKind p) {
            if (!type.isAnnotated()) {
                if (GUTChecker.isAnyDefault(type)) {
                    type.addAnnotation(gchecker.ANY);
                } else {
                    type.addAnnotation(gchecker.PEER);
                }
            }
            return super.visitArray(type, p);
        }
    }
}
