package GUTI

import checkers.inference._
import checkers.inference.pbssolver._

import GUT.GUTChecker

import checkers.util.AnnotationUtils
import scala.collection.mutable.HashMap
import com.sun.source.tree.Tree.Kind

object GUTConstants {
  // call getRealChecker to ensure that it is initialized
  val rc = InferenceMain.getRealChecker.asInstanceOf[GUTChecker]
  val BOTTOM = Constant(rc.BOTTOM)
  val SELF = Constant(rc.SELF)
  val PEER = Constant(rc.PEER)
  val REP = Constant(rc.REP)
  val LOST = Constant(rc.LOST)
  val ANY = Constant(rc.ANY)
}

class GUTIConstraintSolver extends PBSConstraintSolver {

  import GUTConstants._

  val NRBITS = 4
  val CONSTANTS = List(ANY, PEER, REP, LOST)

  // Map from bits to the corresponding variable and constant.
  // There is no mapping for FormulaVariables.
  val bitmap = new HashMap[Int, (AbstractVariable, Constant)]

  def bit(v: AbstractVariable, c: Constant): String = {
    // add 1, because zero is special in CNF
    val base = v.id * 4 + 1

    val bitid = c match {
      case PEER => base + 0
      case REP => base + 1
      case ANY => base + 2
      case LOST => base + 3
    }
    bitmap += bitid -> (v, c)

    bitid.toString
  }

  private def bvarBase = (variables.size + combvariables.size) * 4 + 1

  def bvarBit(bv: BVar): String = {
    (bvarBase + bv.id).toString
  }

  def unbit(in: Int): Elem = {
    if (in < 0) {
      Not(unbit(in.abs))
    } else {
      val (v, c) = bitmap(in)

      if (c != null) {
        EqConst(v, c)
      } else {
        BVar(in - bvarBase)
      }
    }
  }

  def decl(v: AbstractVariable): (AndOfOrs, List[BVar]) = {
    if (v.isInstanceOf[CombVariable]) {
      // lost is allowed for these internal variables
      val o1: OrOfElems = EqConst(v, ANY).or(EqConst(v, PEER)).or(EqConst(v, REP)).or(EqConst(v, LOST))
      val o2 = Not(EqConst(v, PEER)).or(Not(EqConst(v, REP)))
      val o3 = Not(EqConst(v, PEER)).or(Not(EqConst(v, ANY)))
      val o4 = Not(EqConst(v, ANY)).or(Not(EqConst(v, REP)))
      val o5 = Not(EqConst(v, LOST)).or(Not(EqConst(v, REP)))
      val o6 = Not(EqConst(v, LOST)).or(Not(EqConst(v, ANY)))
      val o7 = Not(EqConst(v, LOST)).or(Not(EqConst(v, PEER)))

      (AndOfOrs(o1, o2, o3, o4, o5, o6, o7), null)

    } else {
      // variables that will be inserted into the program cannot be lost
      val o1: OrOfElems = EqConst(v, ANY).or(EqConst(v, PEER)).or(EqConst(v, REP))
      val o2 = Not(EqConst(v, PEER)).or(Not(EqConst(v, REP)))
      val o3 = Not(EqConst(v, PEER)).or(Not(EqConst(v, ANY)))
      val o4 = Not(EqConst(v, ANY)).or(Not(EqConst(v, REP)))
      val o5 = OrOfElems(Not(EqConst(v, LOST)))
      (AndOfOrs(o1, o2, o3, o4, o5), null)
    }
  }

  def encodeSubtype(sub: Slot, sup: Slot): (AndOfOrs, List[BVar]) = {
    sub match {
      case vsub: AbstractVariable => {
        sup match {
          case vsup: AbstractVariable => {
            val o1 = EqConst(vsub, ANY).implies(EqConst(vsup, ANY))
            val o2 = EqConst(vsup, PEER).implies(EqConst(vsub, PEER))
            val o3 = EqConst(vsup, REP).implies(EqConst(vsub, REP))
            // vsub=LOST => (vsup=ANY \/ vsup=LOST)
            val o4 = Not(EqConst(vsub, LOST)).or(EqConst(vsup, ANY)).or(EqConst(vsup, LOST))
            val and = AndOfOrs(o1, o2, o3, o4)
            (and, null)
          }
          case BOTTOM => {
            println("GUTIConstraintSolver::encodeSubtype between " + vsub + " and " + sup)
            (AndOfOrs(EqConst(vsub, PEER)), null)
          }
          case SELF => {
            println("GUTIConstraintSolver::encodeSubtype between " + vsub + " and " + sup)
            (AndOfOrs(EqConst(vsub, PEER)), null)
          }
          case PEER => {
            (AndOfOrs(EqConst(vsub, PEER)), null)
          }
          case REP => {
            (AndOfOrs(EqConst(vsub, REP)), null)
          }
          case LOST => {
            // if the supertype is LOST, everything but ANY is allowed
            (AndOfOrs(Not(EqConst(vsub, ANY))), null)
          }
          case ANY => {
            // if the supertype is ANY, we don't generate a constraint
            // (AndOfOrs(EqConst(vsub, PEER).or(EqConst(vsub, REP)).or(EqConst(vsub, LOST)).or(EqConst(vsub, ANY))), null)
            (null, null)
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between variable and literal: " + sub + " and " + sup)
            (null, null)
          }
        }
      }
      case LiteralThis => {
        sup match {
          case vsup: AbstractVariable => {
            // If the subtype is the "this" literal, the supertype is PEER or ANY
            (AndOfOrs(EqConst(vsup, PEER).or(EqConst(vsup, ANY))), null)
          }
          case PEER => {
            // Valid subtype relation
            (null, null)
          }
          case LOST => {
            // Valid subtype relation
            (null, null)
          }
          case ANY => {
            // Valid subtype relation
            (null, null)
          }
          case _ => {
            println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between: " + sub + " and " + sup)
            (null, null)
          }
        }
      }
      case LiteralSuper => {
        sup match {
          case vsup: AbstractVariable => {
            // If the subtype is the "super" literal, the supertype is PEER or ANY
            (AndOfOrs(EqConst(vsup, PEER).or(EqConst(vsup, ANY))), null)
          }
          case PEER => {
            // Valid subtype relation
            (null, null)
          }
          case LOST => {
            // Valid subtype relation
            (null, null)
          }
          case ANY => {
            // Valid subtype relation
            (null, null)
          }
          case _ => {
            println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between: " + sub + " and " + sup)
            (null, null)
          }
        }
      }
      case LiteralNull => {
        // TODO: anything special to add?
        (null, null)
      }
      case lsub: Literal => {
        sup match {
          case vsup: AbstractVariable => {
            // TODO: only for String literals, see Equality constraints
            // If the subtype is a literal, the supertype is PEER or ANY,
            // because we currently use PEER as default for Strings etc.
            (AndOfOrs(EqConst(vsup, PEER).or(EqConst(vsup, ANY))), null)
          }
          case lsup: AbstractLiteral => {
            println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between two literals: " + sub + " and " + sup)
            (null, null)
          }
          case csup: Constant => {
            // TODO: happens e.g. for null literal, make cases finer
            // println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between literal and constant: " + sub + " and " + sup)
            (null, null)
          }
        }
      }
      case csub: Constant => {
        sup match {
          case vsup: AbstractVariable => {
            csub match {
              case BOTTOM => {
                // subtype of everthing
                (null, null)
              }
              case SELF => {
                (AndOfOrs(EqConst(vsup, PEER).or(EqConst(vsup, ANY))), null)
              }
              case PEER => {
                // ignore LOST as supertype?
                (AndOfOrs(EqConst(vsup, PEER).or(EqConst(vsup, ANY))), null)
              }
              case REP => {
                // ignore LOST as supertype?
                (AndOfOrs(EqConst(vsup, REP).or(EqConst(vsup, ANY))), null)
              }
              case LOST => {
                // ignore LOST as supertype?
                (AndOfOrs(EqConst(vsup, LOST).or(EqConst(vsup, ANY))), null)
              }
              case ANY => {
                // ignore LOST as supertype?
                (AndOfOrs(EqConst(vsup, ANY)), null)
              }
            }
          }
          case lsup: AbstractLiteral => {
            println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between constant and literal: " + sub + " and " + sup)
            (null, null)
          }
          case csup: Constant => {
            if (csub == csup || csup == ANY || ((csub == PEER || csub == REP || csub == SELF) && csup == LOST) ||
                (csub == SELF && (csup == PEER || csup == ANY)) ||
                csub == BOTTOM) {
              // nothing todo, valid subtypes
            } else {
              println("GUTConstraintSolver::encodeSubtype: Something is wrong! Subtype constraint between two constants: " + sub + " and " + sup)
            }
            (null, null)
          }
        }
      }
    }
  }

  def encodeCombine(target: Slot, decl: Slot, res: Slot): (AndOfOrs, List[BVar]) = {
    if (!res.isInstanceOf[CombVariable]) {
      println("encodeCombine should only be called with a CombVariable as third argument!")
      return (null, null)
    }
    val vres = res.asInstanceOf[AbstractVariable]

    target match {
      case vtarget: AbstractVariable => {
        decl match {
          case vdecl: AbstractVariable => {
            // (A /\ B) => C
            // !(A /\ B) \/ C
            // !A \/ !B \/ C
            val o1 = Not(EqConst(vtarget, PEER)).or(Not(EqConst(vdecl, PEER))).or(EqConst(vres, PEER))
            val o2 = Not(EqConst(vtarget, REP)).or(Not(EqConst(vdecl, PEER))).or(EqConst(vres, REP))
            val o3 = Not(EqConst(vdecl, ANY)).or(EqConst(vres, ANY))
            val o4 = Not(EqConst(vdecl, LOST)).or(EqConst(vres, LOST))
            // (A /\ !B) => C
            val o5 = Not(EqConst(vtarget, ANY)).or(EqConst(vdecl, ANY)).or(EqConst(vres, LOST))
            val o6 = Not(EqConst(vtarget, LOST)).or(EqConst(vdecl, ANY)).or(EqConst(vres, LOST))
            // Access on REP only allowed on LiteralThis
            val o7 = Not(EqConst(vdecl, REP)).or(EqConst(vres, LOST))
            val and = AndOfOrs(o1, o2, o3, o4, o5, o6, o7)
            (and, null)
          }
          case cdecl: Constant => {
            cdecl match {
              case SELF => {
                // Self is used as main modifier of extends and implements clauses
                // Treat SELF and PEER the same
                val o1 = Not(EqConst(vtarget, PEER)).or(EqConst(vres, PEER))
                val o2 = Not(EqConst(vtarget, REP)).or(EqConst(vres, REP))
                val o3 = Not(EqConst(vtarget, ANY)).or(EqConst(vres, LOST))
                val o4 = Not(EqConst(vtarget, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case PEER => {
                val o1 = Not(EqConst(vtarget, PEER)).or(EqConst(vres, PEER))
                val o2 = Not(EqConst(vtarget, REP)).or(EqConst(vres, REP))
                val o3 = Not(EqConst(vtarget, ANY)).or(EqConst(vres, LOST))
                val o4 = Not(EqConst(vtarget, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case REP => {
                // note that access through "this" is not covered in
                // this branch, as "this" is a literal
                (AndOfOrs(EqConst(vres, LOST)), null)
              }
              case ANY => {
                (AndOfOrs(EqConst(vres, ANY)), null)
              }
              case LOST => {
                (AndOfOrs(EqConst(vres, LOST)), null)
              }
              case _ => {
                println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between: " + target + " and " + decl)
                (null, null)
              }
            }
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between variable and literal: " + target + " and " + decl)
            (null, null)
          }
        }
      }
      case LiteralThis => {
        decl match {
          case vdecl: AbstractVariable => {
            // this |> v = v
            (equalVars(vres, vdecl), null)
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between two literals: " + target + " and " + decl)
            (null, null)
          }
          case cdecl: Constant => {
            // this |> c = c
            (AndOfOrs(EqConst(vres, cdecl)), null)
          }
        }
      }
      case LiteralSuper => {
        decl match {
          case vdecl: AbstractVariable => {
            // super |> v = v
            (equalVars(vres, vdecl), null)
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between two literals: " + target + " and " + decl)
            (null, null)
          }
          case cdecl: Constant => {
            // super |> c = c
            (AndOfOrs(EqConst(vres, cdecl)), null)
          }
        }
      }
      case LiteralNull => {
        // TODO: anything special to add?
        (null, null)
      }
      case ltarget: Literal => {
        decl match {
          case vdecl: AbstractVariable => {
            // TODO
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between var/literal: " + target + " and " + decl)

            (null, null)
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between two literals: " + target + " and " + decl)
            (null, null)
          }
          case cdecl: Constant => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between literal and constant: " + target + " and " + decl)
            (null, null)
          }
        }
      }
      case ctarget: Constant => {
        decl match {
          case vdecl: AbstractVariable => {
            ctarget match {
              case SELF => {
                (equalVars(vdecl, vres), null)
              }
              case PEER => {
                val o1 = Not(EqConst(vdecl, PEER)).or(EqConst(vres, PEER))
                val o2 = Not(EqConst(vdecl, REP)).or(EqConst(vres, LOST))
                val o3 = Not(EqConst(vdecl, ANY)).or(EqConst(vres, ANY))
                val o4 = Not(EqConst(vdecl, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case REP => {
                val o1 = Not(EqConst(vdecl, PEER)).or(EqConst(vres, REP))
                val o2 = Not(EqConst(vdecl, REP)).or(EqConst(vres, LOST))
                val o3 = Not(EqConst(vdecl, ANY)).or(EqConst(vres, ANY))
                val o4 = Not(EqConst(vdecl, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case ANY => {
                val o1 = Not(EqConst(vdecl, PEER)).or(EqConst(vres, LOST))
                val o2 = Not(EqConst(vdecl, REP)).or(EqConst(vres, LOST))
                val o3 = Not(EqConst(vdecl, ANY)).or(EqConst(vres, ANY))
                val o4 = Not(EqConst(vdecl, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case LOST => {
                val o1 = Not(EqConst(vdecl, PEER)).or(EqConst(vres, LOST))
                val o2 = Not(EqConst(vdecl, REP)).or(EqConst(vres, LOST))
                val o3 = Not(EqConst(vdecl, ANY)).or(EqConst(vres, ANY))
                val o4 = Not(EqConst(vdecl, LOST)).or(EqConst(vres, LOST))
                val and = AndOfOrs(o1, o2, o3, o4)
                (and, null)
              }
              case _ => {
                println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between: " + target + " and " + decl)
                (null, null)
              }
            }
          }
          case ldecl: AbstractLiteral => {
            println("GUTConstraintSolver::encodeCombine: Something is wrong! Combine constraint between constant and literal: " + target + " and " + decl)
            (null, null)
          }
          case cdecl: Constant => {
            if (ctarget==SELF) {
              (AndOfOrs(EqConst(vres, cdecl)), null)
            } else if (ctarget == PEER && cdecl == PEER) {
              (AndOfOrs(EqConst(vres, PEER)), null)
            } else if (ctarget == REP && cdecl == PEER) {
              (AndOfOrs(EqConst(vres, REP)), null)
            } else if (cdecl == ANY) {
              (AndOfOrs(EqConst(vres, ANY)), null)
            } else {
              (AndOfOrs(EqConst(vres, LOST)), null)
            }
          }
        }
      }
    }
  }

  private def equalVars(v1: AbstractVariable, v2: AbstractVariable): AndOfOrs = {
    AndOfOrs(
      (for (c <- CONSTANTS) yield {
        // v1=c => v2=c
        (Not(EqConst(v1, c)).or(EqConst(v2, c)))
      }): _*)
  }

  def encodeEquality(ell: Slot, elr: Slot): (AndOfOrs, List[BVar]) = {
    ell match {
      case vell: AbstractVariable => {
        elr match {
          case velr: AbstractVariable => {
            val o1 = EqConst(vell, ANY).implies(EqConst(velr, ANY))
            val o2 = EqConst(vell, PEER).implies(EqConst(velr, PEER))
            val o3 = EqConst(vell, REP).implies(EqConst(velr, REP))
            val o4 = EqConst(vell, LOST).implies(EqConst(velr, LOST))
            val and = AndOfOrs(o1, o2, o3, o4)
            (and, null)
          }
          case celr: Constant => {
            (AndOfOrs(EqConst(vell, celr)), null)
          }
          case Literal(Kind.STRING_LITERAL, _) => {
            // TODO: currently we use PEER as default for Strings
            (AndOfOrs(EqConst(vell, PEER)), null)
          }
          case ldecl: AbstractLiteral => {
            println("Something is wrong! Equality constraint between variable and literal: " + ell + " and " + elr)
            (null, null)
          }
        }
      }
      case Literal(Kind.STRING_LITERAL, _) => {
        elr match {
          case velr: AbstractVariable => {
            // TODO: currently we use PEER as default for Strings
            (AndOfOrs(EqConst(velr, PEER)), null)
          }
          case lelr: AbstractLiteral => {
            // nothing to do?
            (null, null)
          }
          case PEER => {
            // these happen when PEER is the default for Strings. TODO
            (null, null)
          }
          case celr: Constant => {
            println("GUTConstraintSolver::encodeEqualityConstraint: Something is wrong! Equality constraint between literal and constant: " + ell + " and " + elr)
            (null, null)
          }
        }
      }
      case lell: AbstractLiteral => {
        println("Something is wrong! Equality constraint between literal and something: " + ell + " and " + elr)
        (null, null)
      }
      case cell: Constant => {
        elr match {
          case velr: AbstractVariable => {
            (AndOfOrs(EqConst(velr, cell)), null)
          }
          case lelr: AbstractLiteral => {
            println("Something is wrong! Equality constraint between constant and literal: " + ell + " and " + elr)
            (null, null)
          }
          case celr: Constant => {
            if (cell != celr) {
              println("Something is wrong! Equality constraint between two constants: " + ell + " and " + elr)
            } else {
              // The two constants are equal, so the constraint is satisfied
            }
            (null, null)
          }
        }
      }
    }
  }

  def encodeInequality(ell: Slot, elr: Slot): (AndOfOrs, List[BVar]) = {
    ell match {
      case vell: AbstractVariable => {
        elr match {
          case velr: AbstractVariable => {
            val o1 = EqConst(vell, ANY).implies(Not(EqConst(velr, ANY)))
            val o2 = EqConst(vell, PEER).implies(Not(EqConst(velr, PEER)))
            val o3 = EqConst(vell, REP).implies(Not(EqConst(velr, REP)))
            val o4 = EqConst(vell, LOST).implies(Not(EqConst(velr, LOST)))
            val and = AndOfOrs(o1, o2, o3, o4)
            (and, null)
          }
          case celr: Constant => {
            (AndOfOrs(Not(EqConst(vell, celr))), null)
          }
          case ldecl: AbstractLiteral => {
            println("Something is wrong! Inequality constraint between variable and literal: " + ell + " and " + elr)
            (null, null)
          }
        }
      }
      case lell: AbstractLiteral => {
        println("Something is wrong! Inequality constraint between literal and something: " + ell + " and " + elr)
        (null, null)
      }
      case cell: Constant => {
        elr match {
          case velr: AbstractVariable => {
            (AndOfOrs(Not(EqConst(velr, cell))), null)
          }
          case lelr: AbstractLiteral => {
            println("Something is wrong! Inequality constraint between constant and literal: " + ell + " and " + elr)
            (null, null)
          }
          case celr: Constant => {
            if (cell == celr) {
              println("Something is wrong! Inequality constraint between two constants: " + ell + " and " + elr)
            } else {
              // The two constants are not equal, so the constraint is satisfied
            }
            (null, null)
          }
        }
      }
    }
  }

  def encodeComparable(ell: Slot, elr: Slot): (AndOfOrs, List[BVar]) = {
    ell match {
      case vell: AbstractVariable => {
        elr match {
          case velr: AbstractVariable => {
            (AndOfOrs(EqConst(vell, PEER).implies(Not(EqConst(velr, REP))),
              EqConst(vell, REP).implies(Not(EqConst(velr, PEER)))), null)
          }
          case PEER => {
            (AndOfOrs(EqConst(vell, PEER).or(EqConst(vell, ANY))), null)
          }
          case REP => {
            (AndOfOrs(EqConst(vell, REP).or(EqConst(vell, ANY))), null)
          }
          case LOST => {
            // if one of the sides is LOST or ANY, we don't need a constraint
            (null, null)
          }
          case ANY => {
            (null, null)
          }
          case lelr: AbstractLiteral => {
            // TODO: handle Strings and null as special cases that are OK
            // println("Something is wrong! Comparable constraint between variable and literal: " + ell + " and " + elr)
            (null, null)
          }
        }
      }
      case cell: Constant => {
        elr match {
          case velr: AbstractVariable => {
            cell match {
              case PEER => {
                (AndOfOrs(EqConst(velr, PEER).or(EqConst(velr, ANY))), null)
              }
              case REP => {
                (AndOfOrs(EqConst(velr, REP).or(EqConst(velr, ANY))), null)
              }
              case LOST => {
                // if one of the sides is LOST or ANY, we don't need a constraint
                (null, null)
              }
              case ANY => {
                (null, null)
              }
            }
          }
          case lelr: AbstractLiteral => {
            println("Something is wrong! Comparable constraint between constant and literal: " + ell + " and " + elr)
            (null, null)
          }
          case celr: Constant => {
            println("Something is wrong! Comparable constraint between constant and constant: " + ell + " and " + elr)
            (null, null)
          }
        }
      }
      case lell: AbstractLiteral => {
        println("Something is wrong! Comparable constraint between literal and something: " + ell + " and " + elr)
        (null, null)
      }
    }

    /* This seems too complicated:
       TODO: can we get rid of the BVar everywhere?
    val (oneAnds, oneBs) = encodeSubtype(ell, elr)
    val (twoAnds, twoBs) = encodeSubtype(elr, ell)
    val (resAnds, resB) = oneAnds or twoAnds
    (resAnds, List(resB))
    */
  }

}
